from parquetranger import TableRepo

from .test_core import df1, df2


def test_envs(tmp_path):

    defp = tmp_path / "def"
    testp = tmp_path / "test"

    trepo = TableRepo(defp / "data", env_parents={"test": testp})

    trepo.extend(df1)
    trepo.set_env("test")
    trepo.extend(df2)

    assert trepo.get_full_df().equals(df2)
    trepo.set_env_to_default()
    assert trepo.get_full_df().equals(df1)


def test_env_ctx(tmp_path):

    fp1 = tmp_path / "fp1"
    fp2 = tmp_path / "fp2"

    trepo = TableRepo(fp1 / "data", env_parents={"e1": fp1, "e2": fp2})

    assert trepo._current_env == "e1"

    def _as1():
        assert trepo.get_full_df().equals(df1)

    def _as2():
        assert trepo.get_full_df().equals(df2)

    trepo.extend(df1)
    with trepo.env_ctx("e2"):
        trepo.extend(df2)

    _as1()
    trepo.set_env("e2")
    _as2()
    with trepo.env_ctx("e1"):
        _as1()
    _as2()

    trepo.set_env_to_default()
    _as1()



from functools import partial
from itertools import product

import dask.dataframe as dd
import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from parquetranger import TableRepo


@pytest.mark.parametrize(
    ["seed", "batches", "rowcount", "max_records", "group_cols"],
    product([2, 3], [5, 10, 50], [10, 100, 1000], [0, 90, 900], [None, "C"]),
)
def test_para_extend(
    tmp_path, seed, batches, rowcount, max_records, group_cols, dask_client
):
    rng = np.random.RandomState(seed)
    test_dfs = [
        pd.DataFrame(
            {
                "A": rng.rand(rowcount),
                "B": rng.rand(rowcount),
                "C": rng.randint(100, 105, size=rowcount),
            }
        )
        for _ in range(batches)
    ]

    def _fit(_df):
        return _df.sort_values(["A", "B"]).reset_index(drop=True)

    test_df = pd.concat(test_dfs).pipe(_fit)

    trepo1 = TableRepo(
        tmp_path / "data-pmap",
        group_cols=group_cols,
        max_records=max_records,
        dask_client_address=dask_client.scheduler.address,
    )
    trepo1.batch_extend(test_dfs)
    assert_frame_equal(trepo1.get_full_df().pipe(_fit), test_df)

    for nparts in [3, 10, 30]:
        ddf = dd.from_pandas(test_df, npartitions=nparts)
        trepo2 = TableRepo(
            tmp_path / f"data-{nparts}-daskparts",
            group_cols=group_cols,
            max_records=max_records,
            dask_client_address=dask_client.scheduler.address,
        )
        trepo2.extend(ddf)
        assert_frame_equal(test_df, trepo2.get_full_df().pipe(_fit))


def test_map_partitions_extend(tmp_path):

    seed = 100
    rowcount = 10_000
    npartitions = 10
    group_cols = ["C"]
    max_records = 1000

    rng = np.random.RandomState(seed)
    troot = tmp_path / "data"
    trepo1 = TableRepo(
        troot,
        group_cols=group_cols,
        max_records=max_records,
    )

    df = pd.DataFrame(
        {
            "A": rng.rand(rowcount),
            "B": rng.rand(rowcount),
            "C": rng.randint(100, 105, size=rowcount),
        },
        index=pd.Series(range(0, rowcount)).astype(str).str.zfill(10),
    )

    ddf = dd.from_pandas(df, npartitions=npartitions)
    trepo1.extend(ddf)
    assert_frame_equal(df, trepo1.get_full_df().sort_index())


@pytest.mark.parametrize(
    ["rowcount", "max_records", "group_cols"],
    product([10, 100, 1000], [0, 90, 900], ["C", ["C", "G"]]),
)
def test_native_map_partitions(tmp_path, rowcount, max_records, group_cols):

    seed = 100

    rng = np.random.RandomState(seed)
    trepo1 = TableRepo(
        tmp_path / "d1",
        group_cols=group_cols,
        max_records=max_records,
    )

    trepo2 = TableRepo(
        tmp_path / "d2",
    )

    df = pd.DataFrame(
        {
            "A": rng.rand(rowcount),
            "B": rng.rand(rowcount),
            "C": rng.randint(100, 105, size=rowcount).astype(float),
            "G": rng.choice(["x", "y", "z"], size=rowcount),
        },
        index=pd.Series(range(0, rowcount)).astype(str).str.zfill(10),
    )
    trepo1.extend(df)
    trepo1.map_partitions(partial(_gbmapper, trepo=trepo2, gcols=group_cols))
    assert_frame_equal(
        df.groupby(group_cols)[["A", "B"]].mean().reset_index(),
        trepo2.get_full_df().sort_values(group_cols).reset_index(drop=True),
    )


def _gbmapper(gdf, trepo, gcols):
    gdf.groupby(gcols)[["A", "B"]].mean().reset_index().pipe(
        trepo.extend, try_dask=False
    )



import pandas as pd
import pytest

from parquetranger import TableRepo


@pytest.mark.parametrize(
    ["indices"],
    [
        (None,),
        (pd.Series([1, 2], name="fing"),),
        (pd.MultiIndex.from_product([["A", "C"], [1]], names=["ix", "iy"]),),
    ],
)
def test_diff_cols(tmp_path, indices):

    _df1 = pd.DataFrame({"A": [1, 2], "C": ["g1", "g1"]}, index=indices)
    _df2 = pd.DataFrame({"B": [1, 2], "C": ["g2", "g2"]}, index=indices)

    trepo = TableRepo(tmp_path / "diffcols", group_cols="C", ensure_same_cols=True)
    trepo.extend(_df1)
    trepo.extend(_df2)

    df = trepo.get_full_df()
    assert _df1.columns.union(_df2.columns).isin(df.columns).all()


def test_diff_schema(tmp_path):

    _df1 = pd.DataFrame({"A": [1, 2], "C": ["g1", "g1"]})
    _df2 = pd.DataFrame({"A": [1.2, 2.2], "C": ["g2", "g2"]})

    trepo = TableRepo(tmp_path / "diffcols", group_cols="C", ensure_same_cols=True)
    trepo.extend(_df1)
    trepo.extend(_df2)

    df = trepo.get_full_df()
    assert _df1.columns.union(_df2.columns).isin(df.columns).all()



from itertools import product
from pathlib import Path

import dask.dataframe as dd
import pandas as pd
import pytest

from parquetranger import TableRepo
from parquetranger.core import EXTENSION

df1 = pd.DataFrame(
    {
        "A": [1, 2, 3],
        "B": ["x", "y", "z"],
        "C": [1, 2, 1],
        "C2": ["a", "a", "b"],
    },
    index=["a1", "a2", "a3"],
)
df2 = pd.DataFrame(
    {
        "A": [5, 6, 7],
        "B": ["c", "d", "e"],
        "C": [2, 1, 2],
        "C2": ["b", "b", "a"],
    },
    index=["b1", "b2", "b3"],
)

df3 = pd.DataFrame(
    {
        "A": [9, 3, 1],
        "B": ["f", "g", "h"],
        "C": [2, 1, 3],
        "C2": ["a", "a", "a"],
    },
    index=["c1", "c2", "c3"],
)

df4 = pd.DataFrame(
    {
        "A": [10, 9, 17],
        "B": ["x", "da", "ex"],
        "C": [2, 1, 3],
        "C2": ["ba", "b", "a"],
    },
    index=["b4", "b2", "b3"],
)


@pytest.mark.parametrize(
    ["gb_cols"],
    [
        ("C",),
        (["C", "C2"],),
        (["C2", "C"],),
    ],
)
def test_groupby(tmp_path, gb_cols):

    troot = tmp_path / "data"
    trepo = TableRepo(troot, group_cols=gb_cols)
    base = []
    for _df in [df1, df2, df3]:
        trepo.extend(_df)
        base.append(_df)
        for gid, gdf in pd.concat(base).groupby(gb_cols):
            if not isinstance(gid, tuple):
                gid = (gid,)
            gpath = Path(troot, *map(str, gid)).with_suffix(EXTENSION)
            assert gdf.equals(pd.read_parquet(gpath))
            assert any(map(gdf.equals, trepo.dfs))
        conc = pd.concat(base)
        full_df = trepo.get_full_df()
        assert conc.reindex(full_df.index).equals(full_df)


def test_gb_maxrecs(tmp_path):
    troot = tmp_path / "data"
    trepo = TableRepo(troot, group_cols="C2", max_records=2)
    trepo.extend(df1)
    assert len(trepo.paths) == 2
    trepo.extend(df2)
    assert len(trepo.paths) == 4
    full_df = trepo.get_full_df()
    assert pd.concat([df1, df2]).reindex(full_df.index).equals(full_df)


@pytest.mark.parametrize(
    ["max_records", "n_files"],
    [
        (0, 1),
        (10, 1),
        (5, 2),
        (3, 3),
    ],
)
def test_extender_records(tmp_path, max_records, n_files):

    trepo = TableRepo(tmp_path, max_records=max_records)
    base = []
    for _df in [df1, df2, df3]:
        trepo.extend(_df)
        base.append(_df)

        conc = pd.concat(base)
        full_df = trepo.get_full_df()
        assert conc.reindex(full_df.index).equals(full_df)

    assert trepo.n_files == n_files


@pytest.mark.parametrize(
    ["max_records", "n_files"],
    [
        (0, 1),
        (10, 1),
        (2, 2),
    ],
)
def test_replace_records(tmp_path, max_records, n_files):
    trepo = TableRepo(tmp_path, max_records=max_records)
    trepo.replace_records(df2)

    full_df = trepo.get_full_df()
    assert df2.reindex(full_df.index).equals(full_df)

    trepo.replace_records(df4)

    full_df = trepo.get_full_df()
    new_df = df2.drop(df4.index, errors="ignore").append(df4)
    assert new_df.reindex(full_df.index).equals(full_df)
    assert trepo.n_files == n_files

    dupind_df = pd.DataFrame(
        {"A": [2, 1], "B": ["1", "1"], "C": [1, 1], "C2": ["1", "1"]},
        index=["b2", "b2"],
    )
    trepo.replace_records(dupind_df)
    assert full_df.shape[0] == trepo.get_full_df().shape[0]


def test_gb_replace(tmp_path, dask_client):

    _df1 = pd.DataFrame(
        {
            "A": [1, 2, 3],
            "B": ["x", "x", "y"],
        },
        index=["x1", "x2", "y1"],
    )

    _df2 = pd.DataFrame(
        {
            "A": [10, 20],
            "B": ["x", "y"],
        },
        index=["x1", "y2"],
    )

    _mdf = pd.DataFrame(
        {"A": [10, 2, 3, 20], "B": ["x", "x", "y", "y"]},
        index=["x1", "x2", "y1", "y2"],
    )

    trepo = TableRepo(
        tmp_path,
        group_cols="B",
        dask_client_address=dask_client.scheduler.address,
    )
    trepo.replace_records(_df1)
    assert _df1.equals(trepo.get_full_df())

    trepo.replace_records(_df2, by_groups=True)
    assert _mdf.equals(trepo.get_full_df().sort_index())

    trepo.replace_groups(_df1)
    assert _df1.equals(trepo.get_full_df().sort_index())

    trepo.replace_records(dd.from_pandas(_df2, npartitions=1))
    assert _mdf.equals(trepo.get_full_df().sort_index())


def test_bygroups_error(tmp_path):
    trepo = TableRepo(tmp_path / "fing")
    with pytest.raises(TypeError):
        trepo.replace_records(df1, by_groups=True)

    with pytest.raises(TypeError):
        trepo.replace_groups(df1)


def test_strin(tmp_path):
    _basetest(TableRepo(str(tmp_path / "data" / "subdir")))


@pytest.mark.parametrize(
    ["max_recs", "partitions"],
    [
        (1, 1),
        (0, 2),
        (1, 2),
        (0, 1),
    ],
)
def test_ddf(tmp_path, max_recs, partitions):
    base = []
    trepo = TableRepo(tmp_path / "data", max_recs)
    for _df in [df1, df2]:
        trepo.extend(dd.from_pandas(_df, npartitions=partitions))
        base.append(_df)
        conc = pd.concat(base)
        full_df = trepo.get_full_df()
        assert conc.reindex(full_df.index).equals(full_df)
        assert (
            dd.read_parquet(trepo.full_path).compute().reindex(conc.index).equals(conc)
        )


@pytest.mark.parametrize(
    ["recs"],
    [
        (10,),
        (0,),
    ],
)
def test_ddf_gb(tmp_path, recs):
    base = []
    troot = tmp_path / "fing"
    trepo = TableRepo(troot, recs, group_cols="C")
    for _df in [df1, df2]:
        trepo.extend(dd.from_pandas(_df, npartitions=1))
        base.append(_df)
        conc = pd.concat(base)
        full_df = trepo.get_full_df()
        assert conc.reindex(full_df.index).equals(full_df)
        for gid, gdf in pd.concat(base).groupby("C"):
            pend = (str(gid),)
            if recs:
                pend = (str(gid), "file-{:020d}".format(1))
            gpath = Path(troot, *pend).with_suffix(EXTENSION)
            assert gdf.equals(pd.read_parquet(gpath).reindex(gdf.index))


def test_ddf_empty(tmp_path):
    troot = tmp_path / "fing"
    trepo = TableRepo(troot, group_cols="C")
    ddf = dd.from_pandas(pd.concat([df1, df2]), npartitions=3).loc[
        lambda df: df["A"] > 4, :
    ]
    ddf.pipe(trepo.extend)
    df = ddf.compute()
    assert df.equals(trepo.get_full_df().reindex(df.index))


@pytest.mark.parametrize(
    ["gcols", "max_records"],
    [*product([None, "C", ["C", "C2"], ["C2", "C"], "C2"], [0, 1])],
)
def test_part_paths(tmp_path, gcols, max_records):
    troot = tmp_path / "tab"
    trepo = TableRepo(troot, group_cols=gcols, max_records=max_records)
    trepo.extend(df1)
    if gcols is None:
        with pytest.raises(AttributeError):
            [*trepo.get_partition_paths("C")]
        return

    for part_col in gcols if isinstance(gcols, list) else [gcols]:
        gb_dic = {str(gid): gdf for gid, gdf in df1.groupby(part_col)}
        for gid, gpaths in trepo.get_partition_paths(part_col):
            gdf = dd.read_parquet(gpaths).compute()
            assert gdf.equals(gb_dic[gid])


def _basetest(trepo: TableRepo):
    base = []
    for _df in [df1, df2]:
        trepo.extend(_df)
        base.append(_df)
        conc = pd.concat(base)
        full_df = trepo.get_full_df()
        assert conc.reindex(full_df.index).equals(full_df)
    trepo.replace_all(df3)
    assert trepo.get_full_df().equals(df3)
    trepo.purge()
    assert trepo.get_full_df().empty



import pandas as pd

from parquetranger import TableRepo


def test_simple_metadata(tmp_path):

    df = pd.DataFrame(
        {"d": pd.date_range("2020-01-01", "2020-01-10"), "x": range(2, 12)}
    )

    tpath = tmp_path / "data"

    meta = {"some": "thing", "other": {"thing": ["as", "dict", 10]}}

    trepo = TableRepo(tpath, extra_metadata=meta)
    trepo.replace_all(df)

    full_meta = trepo.full_metadata
    full_meta.pop("pandas")
    assert full_meta == meta
